from pathlib import Path

import pytest
from maggma.stores import JSONStore, MemoryStore
from monty.serialization import dumpfn, loadfn

from emmet.builders.materials.summary import SummaryBuilder
from emmet.builders.vasp.materials import MaterialsBuilder


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(test_dir / "test_si_tasks.json.gz")


@pytest.fixture(scope="session")
def materials(tasks_store):
    materials_store = MemoryStore(key="material_id")
    builder = MaterialsBuilder(tasks=tasks_store, materials=materials_store)
    builder.run()
    return materials_store


@pytest.fixture
def electronic_structure():
    return MemoryStore(key="material_id")


@pytest.fixture
def thermo():
    return MemoryStore(key="material_id")


@pytest.fixture
def grain_boundaries():
    return MemoryStore()


@pytest.fixture
def magnetism():
    return MemoryStore()


@pytest.fixture
def elasticity():
    return MemoryStore()


@pytest.fixture
def dielectric():
    return MemoryStore()


@pytest.fixture
def piezoelectric():
    return MemoryStore()


@pytest.fixture
def phonon():
    return MemoryStore()


@pytest.fixture
def insertion_electrodes():
    return MemoryStore()


@pytest.fixture
def substrates():
    return MemoryStore()


@pytest.fixture
def oxi_states():
    return MemoryStore()


@pytest.fixture
def surfaces():
    return MemoryStore()


@pytest.fixture
def eos():
    return MemoryStore()


@pytest.fixture
def xas():
    return MemoryStore()


@pytest.fixture
def provenance():
    return MemoryStore()


@pytest.fixture
def chemenv():
    return MemoryStore()


@pytest.fixture
def charge_density_index():
    return MemoryStore()


@pytest.fixture
def summary():
    return MemoryStore(key="material_id")


def test_summary_builder(
    materials,
    thermo,
    xas,
    chemenv,
    grain_boundaries,
    electronic_structure,
    magnetism,
    elasticity,
    dielectric,
    piezoelectric,
    phonon,
    insertion_electrodes,
    substrates,
    surfaces,
    oxi_states,
    eos,
    provenance,
    charge_density_index,
    summary,
):

    builder = SummaryBuilder(
        materials=materials,
        electronic_structure=electronic_structure,
        thermo=thermo,
        magnetism=magnetism,
        dielectric=dielectric,
        piezoelectric=piezoelectric,
        phonon=phonon,
        insertion_electrodes=insertion_electrodes,
        elasticity=elasticity,
        substrates=substrates,
        surfaces=surfaces,
        oxi_states=oxi_states,
        xas=xas,
        grain_boundaries=grain_boundaries,
        eos=eos,
        provenance=provenance,
        charge_density_index=charge_density_index,
        summary=summary,
    )

    builder.run()
    assert summary.count() == 1


def test_serialization(tmpdir):
    builder = SummaryBuilder(
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
        MemoryStore(),
    )

    dumpfn(builder.as_dict(), Path(tmpdir) / "test.json")
    loadfn(Path(tmpdir) / "test.json")



import pytest
from maggma.stores import JSONStore, MemoryStore

from emmet.builders.qchem.molecules import MoleculesAssociationBuilder, MoleculesBuilder
from emmet.builders.molecules.orbitals import OrbitalBuilder


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(test_dir / "C3H4Li1O3.json.gz")


@pytest.fixture(scope="session")
def mol_store(tasks_store):
    assoc_store = MemoryStore(key="molecule_id")
    stage_one = MoleculesAssociationBuilder(tasks=tasks_store, assoc=assoc_store)
    stage_one.run()

    mol_store = MemoryStore(key="molecule_id")
    stage_two = MoleculesBuilder(assoc=assoc_store, molecules=mol_store, prefix="libe")
    stage_two.run()

    return mol_store


@pytest.fixture(scope="session")
def orbital_store():
    return MemoryStore()


@pytest.mark.skip(reason="Waiting on molecule update.")
def test_orbital_builder(tasks_store, mol_store, orbital_store):
    builder = OrbitalBuilder(tasks_store, mol_store, orbital_store)
    builder.run()

    assert orbital_store.count() == 27
    assert orbital_store.count({"open_shell": True}) == 14



from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Tuple
from math import ceil
from datetime import datetime

from maggma.core import Builder, Store
from maggma.utils import grouper
from pymatgen.analysis.structure_matcher import ElementComparator, StructureMatcher
from pymatgen.core.structure import Structure

from emmet.builders.settings import EmmetBuildSettings
from emmet.core.provenance import ProvenanceDoc, SNLDict
from emmet.core.utils import get_sg, jsanitize


class ProvenanceBuilder(Builder):
    def __init__(
        self,
        materials: Store,
        provenance: Store,
        source_snls: List[Store],
        settings: Optional[EmmetBuildSettings] = None,
        query: Optional[Dict] = None,
        **kwargs,
    ):
        """
        Creates provenance from source SNLs and materials

        Args:
            materials: Store of materials docs to tag with SNLs
            provenance: Store to update with provenance data
            source_snls: List of locations to grab SNLs
            query : query on materials to limit search
        """
        self.materials = materials
        self.provenance = provenance
        self.source_snls = source_snls
        self.settings = EmmetBuildSettings.autoload(settings)
        self.query = query or {}
        self.kwargs = kwargs

        materials.key = "material_id"
        provenance.key = "material_id"
        for s in source_snls:
            s.key = "snl_id"

        super().__init__(sources=[materials, *source_snls], targets=[provenance], **kwargs)

    def ensure_indicies(self):

        self.materials.ensure_index("material_id", unique=True)
        self.materials.ensure_index("formula_pretty")

        self.provenance.ensure_index("material_id", unique=True)
        self.provenance.ensure_index("formula_pretty")

        for s in self.source_snls:
            s.ensure_index("snl_id")
            s.ensure_index("formula_pretty")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        self.ensure_indicies()

        # Find all formulas for materials that have been updated since this
        # builder was last ran
        q = self.query
        updated_materials = self.provenance.newer_in(self.materials, criteria=q, exhaustive=True)
        forms_to_update = set(self.materials.distinct("formula_pretty", {"material_id": {"$in": updated_materials}}))

        # Find all new SNL formulas since the builder was last run
        for source in self.source_snls:
            new_snls = self.provenance.newer_in(source)
            forms_to_update |= set(source.distinct("formula_pretty", {source.key: {"$in": new_snls}}))

        # Now reduce to the set of formulas we actually have
        forms_avail = set(self.materials.distinct("formula_pretty", self.query))
        forms_to_update = forms_to_update & forms_avail

        mat_ids = set(self.materials.distinct("material_id", {"formula_pretty": {"$in": list(forms_to_update)}})) & set(
            updated_materials
        )

        N = ceil(len(mat_ids) / number_splits)

        self.logger.info(f"Found {len(mat_ids)} new/updated systems to distribute to workers " f"in {N} chunks.")

        for chunk in grouper(mat_ids, N):
            yield {"query": {"material_id": {"$in": chunk}}}

    def get_items(self) -> Tuple[List[Dict], List[Dict]]:  # type: ignore
        """
        Gets all materials to assocaite with SNLs
        Returns:
            generator of materials and SNLs that could match
        """
        self.logger.info("Provenance Builder Started")

        self.logger.info("Setting indexes")
        self.ensure_indicies()

        # Find all formulas for materials that have been updated since this
        # builder was last ran
        q = self.query
        updated_materials = self.provenance.newer_in(self.materials, criteria=q, exhaustive=True)
        forms_to_update = set(self.materials.distinct("formula_pretty", {"material_id": {"$in": updated_materials}}))

        # Find all new SNL formulas since the builder was last run
        for source in self.source_snls:
            new_snls = self.provenance.newer_in(source)
            forms_to_update |= set(source.distinct("formula_pretty", {source.key: {"$in": new_snls}}))

        # Now reduce to the set of formulas we actually have
        forms_avail = set(self.materials.distinct("formula_pretty", self.query))
        forms_to_update = forms_to_update & forms_avail

        mat_ids = set(self.materials.distinct("material_id", {"formula_pretty": {"$in": list(forms_to_update)}})) & set(
            updated_materials
        )

        self.total = len(mat_ids)

        self.logger.info(f"Found {self.total} new/updated systems to process")

        for mat_id in mat_ids:

            mat = self.materials.query_one(
                properties=[
                    "material_id",
                    "last_updated",
                    "structure",
                    "initial_structures",
                    "formula_pretty",
                    "deprecated",
                ],
                criteria={"material_id": mat_id},
            )

            snls = []  # type: list
            for source in self.source_snls:
                snls.extend(source.query(criteria={"formula_pretty": mat["formula_pretty"]}))

            snl_groups = defaultdict(list)
            for snl in snls:
                struc = Structure.from_dict(snl)
                snl_sg = get_sg(struc)
                struc.snl = SNLDict(**snl)
                snl_groups[snl_sg].append(struc)

            mat_sg = get_sg(Structure.from_dict(mat["structure"]))

            snl_structs = snl_groups[mat_sg]

            self.logger.debug(f"Found {len(snl_structs)} potential snls for {mat_id}")
            yield mat, snl_structs

    def process_item(self, item) -> Dict:
        """
        Matches SNLS and Materials
        Args:
            item (tuple): a tuple of materials and snls
        Returns:
            list(dict): a list of collected snls with material ids
        """
        mat, snl_structs = item
        formula_pretty = mat["formula_pretty"]
        snl_doc = None
        self.logger.debug(f"Finding Provenance {formula_pretty}")

        # Match up SNLS with materials

        matched_snls = self.match(snl_structs, mat)

        if len(matched_snls) > 0:
            doc = ProvenanceDoc.from_SNLs(
                material_id=mat["material_id"],
                structure=Structure.from_dict(mat["structure"]),
                snls=matched_snls,
                deprecated=mat["deprecated"],
            )
        else:
            doc = ProvenanceDoc(
                material_id=mat["material_id"],
                structure=Structure.from_dict(mat["structure"]),
                deprecated=mat["deprecated"],
                created_at=datetime.utcnow(),
            )

        doc.authors.append(self.settings.DEFAULT_AUTHOR)
        doc.history.append(self.settings.DEFAULT_HISTORY)
        doc.references.append(self.settings.DEFAULT_REFERENCE)

        snl_doc = jsanitize(doc.dict(exclude_none=False), allow_bson=True)

        return snl_doc

    def match(self, snl_structs, mat):
        """
        Finds a material doc that matches with the given snl
        Args:
            snl_structs ([dict]): the snls struct list
            mat (dict): a materials doc
        Returns:
            generator of materials doc keys
        """

        m_strucs = [Structure.from_dict(mat["structure"])] + [
            Structure.from_dict(init_struc) for init_struc in mat["initial_structures"]
        ]

        sm = StructureMatcher(
            ltol=self.settings.LTOL,
            stol=self.settings.STOL,
            angle_tol=self.settings.ANGLE_TOL,
            primitive_cell=True,
            scale=True,
            attempt_supercell=False,
            allow_subset=False,
            comparator=ElementComparator(),
        )

        snls = []

        for s in m_strucs:
            for snl_struc in snl_structs:
                if sm.fit(s, snl_struc):
                    if snl_struc.snl not in snls:
                        snls.append(snl_struc.snl)

        self.logger.debug(f"Found {len(snls)} SNLs for {mat['material_id']}")
        return snls

    def update_targets(self, items):
        """
        Inserts the new SNL docs into the SNL collection
        """
        snls = list(filter(None, items))

        if len(snls) > 0:
            self.logger.info(f"Found {len(snls)} SNLs to update")
            self.provenance.update(snls)
        else:
            self.logger.info("No items to update")



from pathlib import Path

import pytest
from maggma.stores import JSONStore, MemoryStore
from monty.serialization import dumpfn, loadfn

from emmet.builders.materials.magnetism import MagneticBuilder
from emmet.builders.vasp.materials import MaterialsBuilder


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(test_dir / "magnetism/magnetism_task_docs.json.gz")


@pytest.fixture(scope="session")
def materials_store(tasks_store):
    materials_store = MemoryStore(key="material_id")
    builder = MaterialsBuilder(tasks=tasks_store, materials=materials_store)
    builder.run()
    return materials_store


@pytest.fixture
def magnetism_store():
    return MemoryStore(key="material_id")


def test_magnetism_builder(tasks_store, magnetism_store, materials_store):

    builder = MagneticBuilder(
        tasks=tasks_store, magnetism=magnetism_store, materials=materials_store
    )
    builder.run()

    assert magnetism_store.count() == 4
    assert magnetism_store.count({"deprecated": False}) == 4

    test_mpids = {
        "mp-1289887": "AFM",
        "mp-1369002": "FiM",
        "mp-1791788": "NM",
        "mp-1867075": "FM",
    }

    print(list([doc["material_id"] for doc in magnetism_store.query({})]))

    for mpid in test_mpids:
        doc = magnetism_store.query_one({"material_id": mpid})
        assert doc["ordering"] == test_mpids[mpid]


def test_serialization(tmpdir):
    builder = MagneticBuilder(MemoryStore(), MemoryStore(), MemoryStore())

    dumpfn(builder.as_dict(), Path(tmpdir) / "test.json")
    loadfn(Path(tmpdir) / "test.json")



import pytest
from maggma.stores import JSONStore, MemoryStore

from emmet.builders.qchem.molecules import MoleculesAssociationBuilder, MoleculesBuilder


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(test_dir / "C3H4Li1O3.json.gz")


@pytest.fixture(scope="session")
def assoc_store():
    return MemoryStore()


@pytest.fixture(scope="session")
def mol_store():
    return MemoryStore()

@pytest.mark.xfail(reason="Waiting on molecule updates")
def test_molecules_builder(tasks_store, assoc_store, mol_store):
    stage_one = MoleculesAssociationBuilder(tasks=tasks_store, assoc=assoc_store)
    stage_one.run()
    assert assoc_store.count() == 67
    assert assoc_store.count({"deprecated": True}) == 1

    assoc_store.key = "molecule_id"

    stage_two = MoleculesBuilder(assoc=assoc_store, molecules=mol_store, prefix="libe")
    stage_two.run()

    assert mol_store.count() == 48
    assert mol_store.count({"deprecated": True}) == 0
    assert mol_store.count({"molecule_id": {"$regex": "libe-"}}) == 48



import warnings
from collections import defaultdict
from itertools import chain
from typing import Dict, Iterable, Iterator, List, Optional
from math import ceil

from maggma.core import Builder, Store
from maggma.utils import grouper
from pymatgen.entries.computed_entries import ComputedStructureEntry
from pymatgen.entries.compatibility import Compatibility

from emmet.core.utils import jsanitize
from emmet.builders.utils import chemsys_permutations, HiddenPrints
from emmet.core.thermo import ThermoType
from emmet.core.corrected_entries import CorrectedEntriesDoc


class CorrectedEntriesBuilder(Builder):
    def __init__(
        self,
        materials: Store,
        corrected_entries: Store,
        oxidation_states: Optional[Store] = None,
        query: Optional[Dict] = None,
        compatibility: Optional[List[Compatibility]] = None,
        chunk_size: int = 1000,
        **kwargs,
    ):
        """
        Produces corrected thermo entry data from uncorrected materials entries.
        This is meant to be an intermediate builder for the main thermo builder.

        Args:
            materials (Store): Store of materials documents
            corrected_entries (Store): Store to output corrected entry data
            oxidation_states (Store): Store of oxidation state data to use in correction scheme application
            query (dict): dictionary to limit materials to be analyzed
            compatibility ([Compatability]): Compatability module
                to ensure energies are compatible
            chunk_size (int): Size of chemsys chunks to process at any one time.
        """

        self.materials = materials
        self.query = query if query else {}
        self.corrected_entries = corrected_entries
        self.compatibility = compatibility or [None]
        self.oxidation_states = oxidation_states
        self.chunk_size = chunk_size
        self._entries_cache: Dict[str, List[ComputedStructureEntry]] = defaultdict(list)

        if self.corrected_entries.key != "chemsys":
            warnings.warn(
                "Key for the corrected_entries store is incorrect and has been changed "
                f"from {self.corrected_entries.key} to chemsys!"
            )
            self.corrected_entries.key = "thermo_id"

        if self.materials.key != "material_id":
            warnings.warn(
                f"Key for the materials store is incorrect and has been changed from {self.materials.key} to material_id!"  # noqa: E501
            )
            self.materials.key = "material_id"

        sources = [materials]

        if self.oxidation_states is not None:

            if self.oxidation_states.key != "material_id":
                warnings.warn(
                    f"Key for the oxidation states store is incorrect and has been changed from {self.oxidation_states.key} to material_id!"  # noqa:E501
                )
                self.oxidation_states.key = "material_id"

            sources.append(oxidation_states)  # type: ignore

        targets = [corrected_entries]

        super().__init__(sources=sources, targets=targets, chunk_size=chunk_size, **kwargs)

    def ensure_indexes(self):
        """
        Ensures indicies on the tasks and materials collections
        """

        # Search index for materials
        self.materials.ensure_index("material_id")
        self.materials.ensure_index("chemsys")
        self.materials.ensure_index("last_updated")

        # Search index for corrected_entries
        self.corrected_entries.ensure_index("chemsys")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        to_process_chemsys = self._get_chemsys_to_process()

        N = ceil(len(to_process_chemsys) / number_splits)

        for chemsys_chunk in grouper(to_process_chemsys, N):

            yield {"query": {"chemsys": {"$in": list(chemsys_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets whole chemical systems of entries to process
        """

        self.logger.info("Corrected Entries Builder Started")

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        to_process_chemsys = self._get_chemsys_to_process()

        self.logger.info(f"Processing entries in {len(to_process_chemsys)} chemical systems")
        self.total = len(to_process_chemsys)

        # Yield the chemical systems in order of increasing size
        for chemsys in sorted(to_process_chemsys, key=lambda x: len(x.split("-")), reverse=False):
            entries = self.get_entries(chemsys)
            yield entries

    def process_item(self, item):
        """
        Applies correction schemes to entries and constructs CorrectedEntriesDoc objects
        """

        if not item:
            return None

        entries = [ComputedStructureEntry.from_dict(entry) for entry in item]
        # determine chemsys
        elements = sorted(set([el.symbol for e in entries for el in e.composition.elements]))
        chemsys = "-".join(elements)

        self.logger.debug(f"Processing {len(entries)} entries for {chemsys}")

        all_entry_types = {str(e.data["run_type"]) for e in entries}

        corrected_entries = {}

        for compatability in self.compatibility:

            if compatability:
                if compatability.name == "MP DFT mixing scheme":
                    thermo_type = ThermoType.GGA_GGA_U_R2SCAN
                elif compatability.name == "MP2020":
                    thermo_type = ThermoType.GGA_GGA_U
                else:
                    thermo_type = ThermoType.UNKNOWN

                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    with HiddenPrints():
                        if "R2SCAN" in all_entry_types:
                            pd_entries = compatability.process_entries(entries)

                            only_scan_pd_entries = [e for e in entries if str(e.data["run_type"]) == "R2SCAN"]
                            corrected_entries["R2SCAN"] = only_scan_pd_entries

                        else:
                            pd_entries = compatability.process_entries(entries)

                        corrected_entries[str(thermo_type)] = pd_entries

            else:
                if len(all_entry_types) > 1:
                    raise ValueError("More than one functional type has been provided without a mixing scheme!")
                else:
                    thermo_type = all_entry_types.pop()

                corrected_entries[str(thermo_type)] = entries

        doc = CorrectedEntriesDoc(chemsys=chemsys, entries=corrected_entries)

        return jsanitize(doc.dict(), allow_bson=True)

    def update_targets(self, items):
        """
        Inserts the new corrected entry docs into the corrected entries collection
        """

        docs = list(filter(None, items))

        if len(docs) > 0:
            self.logger.info(f"Updating {len(docs)} corrected entry documents")
            self.corrected_entries.update(docs=docs, key=["chemsys"])
        else:
            self.logger.info("No corrected entry items to update")

    def get_entries(self, chemsys: str) -> List[Dict]:
        """
        Gets entries from the materials collection for the corresponding chemical systems
        Args:
            chemsys (str): a chemical system represented by string elements seperated by a dash (-)
        Returns:
            set (ComputedEntry): a set of entries for this system
        """

        self.logger.info(f"Getting entries for: {chemsys}")
        # First check the cache
        all_chemsys = chemsys_permutations(chemsys)
        cached_chemsys = all_chemsys & set(self._entries_cache.keys())
        query_chemsys = all_chemsys - cached_chemsys
        all_entries = list(chain.from_iterable(self._entries_cache[c] for c in cached_chemsys))

        self.logger.debug(f"Getting {len(cached_chemsys)} sub-chemsys from cache for {chemsys}")
        self.logger.debug(f"Getting {len(query_chemsys)} sub-chemsys from DB for {chemsys}")

        # Second grab the materials docs
        new_q = dict(self.query)
        new_q["chemsys"] = {"$in": list(query_chemsys)}
        new_q["deprecated"] = False

        materials_docs = list(self.materials.query(criteria=new_q, properties=["material_id", "entries", "deprecated"]))

        # Get Oxidation state data for each material
        oxi_states_data = {}
        if self.oxidation_states:
            material_ids = [t["material_id"] for t in materials_docs]
            oxi_states_data = {
                d["material_id"]: d.get("average_oxidation_states", {})
                for d in self.oxidation_states.query(
                    properties=["material_id", "average_oxidation_states"],
                    criteria={"material_id": {"$in": material_ids}, "state": "successful"},
                )
            }

        self.logger.debug(
            f"Got {len(materials_docs)} entries from DB for {len(query_chemsys)} sub-chemsys for {chemsys}"
        )

        # Convert entries into ComputedEntries and store
        for doc in materials_docs:
            for r_type, entry_dict in doc.get("entries", {}).items():
                entry_dict["data"]["oxidation_states"] = oxi_states_data.get(entry_dict["data"]["material_id"], {})
                entry_dict["data"]["run_type"] = r_type
                elsyms = sorted(set([el for el in entry_dict["composition"]]))
                self._entries_cache["-".join(elsyms)].append(entry_dict)
                all_entries.append(entry_dict)

        self.logger.info(f"Total entries in {chemsys} : {len(all_entries)}")

        return all_entries

    def _get_chemsys_to_process(self):
        # Use last-updated to find new chemsys
        materials_chemsys_dates = {}
        for d in self.materials.query(
            {"deprecated": False, **self.query},
            properties=[self.corrected_entries.key, self.materials.last_updated_field]
        ):

            entry = materials_chemsys_dates.get(d[self.corrected_entries.key], None)
            if entry is None or d[self.materials.last_updated_field] > entry:
                materials_chemsys_dates[d[self.corrected_entries.key]] = d[self.materials.last_updated_field]

        corrected_entries_chemsys_dates = {
            d[self.corrected_entries.key]: d[self.corrected_entries.last_updated_field]
            for d in self.corrected_entries.query(
                {}, properties=[self.corrected_entries.key, self.corrected_entries.last_updated_field]
            )
        }

        to_process_chemsys = [
            chemsys
            for chemsys in materials_chemsys_dates
            if (chemsys not in corrected_entries_chemsys_dates)
            or (materials_chemsys_dates[chemsys] > corrected_entries_chemsys_dates[chemsys])
        ]

        return to_process_chemsys
